"""
Module to get connected peers for a given node
by calculating the sending times and choosing the least
sending time peer as a connected peer
:author: gurnoorsingh (20221024)
"""
import re
import os
import json
import subprocess
import time
import logging
from tqdm import tqdm
from RTT import myping
from utils import filter_logs_by_txn_id, get_sender_relays

logging.basicConfig(level=logging.DEBUG, format='[%(lineno)s] - %(levelname)s - %(message)s')


def get_relay_communication_stats(txn_id):
    """
    function to return (sender, send_time) tuples
    for a given transaction id
    """
    send_time_values = dict()
    filename = f"/home/azureuser/algorand-topology/filtered_logs/data_{txn_id}"

    if not os.path.exists(filename):
        logging.warning(f"Log file {filename} not found. Creating it...")
        filter_logs_by_txn_id(txn_id)

    MSG_RECEIVED_IDENTIFIER = "got a tx group with IDs"

    with open(filename) as fh:
        for line in fh.readlines():
            try:
                row = json.loads(line)
            except:
                continue
            if row["msg"].startswith(MSG_RECEIVED_IDENTIFIER):
                txn_ids_str = re.findall(f".*{MSG_RECEIVED_IDENTIFIER} \[(.*?)\].*", row["msg"])[0]
                txn_ids = [x.strip() for x in txn_ids_str.split()]

                assert txn_id in txn_ids

                sender = re.findall(".*rootURL:(.*?)\s.*", row["msg"])[0]
                recv_time = int(re.findall(".*at.*?([0-9]+) with data.*", row["msg"])[0])
                txn_size = len(re.findall(".*uint8=\[(.*?)\]", row["msg"])[0].split())

                try:
                    send_time = recv_time - myping(f"{sender.split(':')[0]}")//2
                    send_time_values[sender] = send_time
                except Exception as e:
                    logging.error(e, sender)
    
    # return dictionary sorted by values
    return dict(sorted(send_time_values.items(), key=lambda x: x[1]))

def custom_sleep(num_seconds):
    iterations = 2 * num_seconds
    for _ in tqdm(list(range(iterations))):
        time.sleep(0.5)

def create_and_send_transaction():
    """
    Function to create a transaction and send it to the network.
    Uses the default bash script to create the transaction, as that
    script is already populated with the required parameters
    :return: returns the string corresponding to the transaction id
    """
    cmd = ["bash", "/home/azureuser/algorand-topology/scripts/create_txn.sh"]
    result = subprocess.run(cmd, stdout=subprocess.PIPE)
    output = result.stdout.decode()
    output_items = output.split('\n')
    txn_id = output_items[2].split(':')[1].strip()
    return txn_id

def restart_node():
    logging.debug("stopping algorand node...")
    subprocess.run(["/home/azureuser/go/bin/goal", "node", "stop", "-d", "/home/azureuser/.algorand"])

    logging.debug("deleting node log file...")
    if os.path.exists("/home/azureuser/.algorand/node.log"):
        os.remove("/home/azureuser/.algorand/node.log")

    restart_cmd = subprocess.run(["/home/azureuser/go/bin/goal", "node", "start", "-d", "/home/azureuser/.algorand"])
    assert restart_cmd.returncode == 0

    log_file_path = "/home/azureuser/.algorand/node.log"
    
    logging.debug("Waiting for node log file to get formed...")
    while not os.path.exists(log_file_path): time.sleep(1)
    logging.debug("Node log file formed")

    sleep_time = 60
    while True:
        grep_cmd = ["grep", "OutgoingPeers", log_file_path] 
        grep_cmd_result = subprocess.run(grep_cmd)
        if grep_cmd_result.returncode == 0:
            break
        logging.warning(f"OutgoingPeers not found, going to sleep for {sleep_time}s...")
        custom_sleep(sleep_time)
        sleep_time = 5


def main():
    # number of times the node should be restarted
    # this is particularly important when the txn creator node
    # is expected to be connected to only 1 relay node i.e. its
    # GossipFanout is set to 1. Restarting will make it connect to
    # different relay node
    num_restarts = 10
    
    for iteration in range(num_restarts):
        logging.info(f"Run{iteration+1}...")

        restart_node()

        txn_num = 100

        # number of times our algorithm predicted the correct connected relay
        exact_match_num = 0

        # number of times our algorithm predicted the correct region
        region_num = 0

        for _ in range(txn_num):
            txn_id = create_and_send_transaction()
            logging.debug(f"Created transaction with id: {txn_id}")
            send_times = get_relay_communication_stats(txn_id)

            sorted_send_times_dict = dict(sorted(send_times.items(), key=lambda x: x[1]))
            predicted_sender_relays = sorted_send_times_dict
                # print(sender, send_time - min_send_time)
            
            # check the ground truth
            true_sender_relays = get_sender_relays()
            
            # for logging purposes
            logging.debug(f"predicted: {predicted_sender_relays} actual: {true_sender_relays}")
    
        


if __name__ == '__main__':
    main()
